#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import httplib
import logging
import json
import re
import time
from datetime import timedelta
import shutil
import optparse

import tornado.ioloop
import tornado.web
import tornado.queues
import tornado.gen

from settings import settings, load_yaml

'''

Essentials:

  * Tests (in progress).
  * Setup.py

Consider:

  * Remove LEGAL_STATUS, allow all status values, leave validation to PXE
  * Alternative persistent storage mechanism.
  * Reset feed when full status is requested by PXE (leaving for integration phase).
  * Use InstallationStorageDummy class for unit testing.

'''

LEGAL_STATUS = {
    'pending_installation',
    'booting_to_installer',
    'installing',
    'booting_to_os',
    'configuring',
    'sign_request',
    'done',
    'error',
    'aborted',
    'unknown',
}

DEFAULT_STATUS = 'unknown'

# Overridden from config:

uuid_validator = re.compile(r'^[0-9a-f][0-9a-f\-]{6,35}$').match
path_validator = re.compile(r'^([\w\d][\w\d_\-\.]*(/[\w\d][\w\d_\-\.]*)?)?$').match


class Installation(object):
    ip = None           # Installation IP, or None if IP is not yet known
    uuid = ''           # Value generated by uuid.uuid4()
    logs = None         # Dict of {filename: uploaded_timestamp}
    files = None        # Dict of {filename: downloaded_timestamp}, or :None if not downloaded yet
    last_log = None     # Name of last uploaded log file or None if no log files are available
    created = None      # Unix timestamp
    updated = None      # Unix timestamp
    inbox = None        # tornado.queues.Queue object to send events from PXE to Installation
    waiting = False     # Indicates if there's currently a `POST /%UUID%` request waiting for event

    session = None      # Session UID that we expect from PXE

    status = DEFAULT_STATUS  # Value from LEGAL_STATUS

    def __init__(self, uuid, status=DEFAULT_STATUS, logs=None, files=None,
                 created=None, updated=None, ip=None):
        self.ip = ip
        self.uuid = uuid
        self.status = status
        if logs:
            self.logs = dict(logs)
            self.last_log = max(logs, key=logs.get)
        else:
            self.logs = dict()
        self.files = dict() if files is None else dict(files)
        self.created = int(time.time()) if created is None else created
        self.updated = int(time.time()) if updated is None else updated
        self.inbox = tornado.queues.Queue(1)

    def full_status(self):
        result = {
            'ip': self.ip,
            'status': self.status,
            'created': self.created,
            'updated': self.updated,
            'downloads': dict(self.files),
        }
        if self.last_log:
            result.update({
                'logfile': self.last_log,
                'logfile_created': self.logs[self.last_log],
            })
        return result

    def update_status(self, status, ip=None):
        self.updated = int(time.time())
        self.status = status
        event = {
            'uuid': self.uuid,
            'status': self.status,
            'updated': self.updated,
        }
        if ip is not None and self.ip is None:
            event['ip'] = self.ip = ip
        return event

    def update_logs(self, logfile, ip=None):
        self.updated = int(time.time())
        self.logs[logfile] = self.updated
        self.last_log = logfile
        event = {
            'uuid': self.uuid,
            'updated': self.updated,
            'logfile': logfile,
            'logfile_created': self.updated,
        }
        if ip is not None and self.ip is None:
            event['ip'] = self.ip = ip
        return event

    def update_downloads(self, file, ip=None):
        self.updated = int(time.time())
        self.files[file] = self.updated
        event = {
            'uuid': self.uuid,
            'updated': self.updated,
            'downloads': {
                file: self.updated,
            }
        }
        if ip is not None and self.ip is None:
            event['ip'] = self.ip = ip
        return event

    def update_new_file(self, file):
        self.updated = int(time.time())
        self.files[file] = None
        return {
            'uuid': self.uuid,
            'updated': self.updated,
            'downloads': {
                file: None,
            }
        }


class InstallationStorageDummy(object):

    logger = None

    def __init__(self, logger):
        self.logger = logger

    def load_installations(self):
        return {}

    def save_installation(self, installations, inst):
        pass


class InstallationStorage(InstallationStorageDummy):
    '''
    Persistent Storage abstraction class.

    Provides two methods. One is called during SmartProxy start to load all existing Installations
    into memory. Second is run to update installation data on disk. It is up to StorageLayer how
    to implement the storage, fully abstracting it from the rest of SmartProxy. Current version
    is storing all data as serialized dictionary inside Installation folder.
    '''

    def load_installations(self):
        installations = {}
        folders = os.listdir(settings.STORAGE_UUID)
        for uuid in folders:
            if not uuid_validator(uuid):
                continue
            folder = os.path.join(settings.STORAGE_UUID, uuid)
            status_file = os.path.join(folder, settings.STORAGE_FILENAME)
            if not os.path.exists(status_file):
                continue
            try:
                with open(status_file, 'rb') as f:
                    serialized = f.read()
            except IOError:
                self.logger.exception('Cannot open status file `%s`', status_file)
                continue
            try:
                data = json.loads(serialized)
            except ValueError:
                self.logger.exception('Cannot unserialize status file `%s`', status_file)
                continue
            logs = data.get('logs', {})
            files = data.get('files', {})
            check_files = set(os.listdir(folder)) - set([settings.STORAGE_FILENAME] + logs.keys())
            for filename in check_files:
                if filename not in files:
                    files[filename] = None
            installations[uuid] = Installation(
                uuid=uuid,
                status=data.get('status', DEFAULT_STATUS),
                logs=logs,
                files=files,
                created=data.get('created', None),
                updated=data.get('updated', None),
                ip=data.get('ip', None),
            )
        return installations

    def save_installation(self, installations, inst):
        if not isinstance(inst, Installation):
            inst = installations[inst]
        folder = os.path.join(settings.STORAGE_UUID, inst.uuid)
        if not os.path.exists(folder):
            os.mkdir(folder)
        status_file = os.path.join(folder, settings.STORAGE_FILENAME)
        data = {
            'status': inst.status,
            'logs': inst.logs,
            'files': inst.files,
            'created': inst.created,
            'updated': inst.updated,
            'ip': inst.ip,
        }
        with open(status_file, 'w+b') as f:
            json.dump(data, f)


class SmartProxyHandler(tornado.web.RequestHandler):

    logger = None
    storage = None
    installations = None
    pxe_feed = None

    def initialize(self, **kwargs):
        if 'logger' in kwargs:
            self.logger = kwargs['logger']
        else:
            self.logger = logging.getLogger(__name__)
        if 'storage' in kwargs:
            self.storage = kwargs['storage']
        else:
            self.storage = InstallationStorageDummy()
        if 'pxe_feed' in kwargs:
            self.pxe_feed = kwargs['pxe_feed']
        else:
            self.pxe_feed = tornado.queues.Queue()
        if 'installations' in kwargs:
            self.installations = kwargs['installations']
        else:
            self.installations = self.storage.load_installations()

    def json_error(self, status_code, message):
        self.set_status(status_code, message)
        self.write({'result': 'error', 'message': message.lower()})


class RootHandler(SmartProxyHandler):

    def get(self, *args, **kwargs):
        if not settings.DEBUG:
            return self.json_error(httplib.FORBIDDEN, 'Debug Mode Only')
        # Manual: Get API reference.
        self.logger.debug('%s.get() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        # Retrieve API reference.
        self.render('smartproxy.html', root='')

    def post(self, *args, **kwargs):
        # PXE: Create new installation
        self.logger.debug('%s.post() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        # Validate installation UUID
        uuid = self.get_body_argument('uuid', None)
        if (uuid is None) or not(uuid_validator(uuid)):
            return self.json_error(httplib.BAD_REQUEST, 'Invalid Request')
        if uuid in self.installations:
            return self.json_error(httplib.CONFLICT, 'Installation Already Exists')
        # Create installation in memory cache.
        self.installations[uuid] = Installation(uuid)
        # Store installation on persistent storage.
        self.storage.save_installation(self.installations, uuid)
        # Successful response.
        self.write({'result': 'ok'})


class ErrorHandler(SmartProxyHandler):

    def get(self, *args, **kwargs):
        self.json_error(httplib.NOT_FOUND, 'Unknown URI Request')

    head = post = delete = patch = put = get


class StatusHandler(SmartProxyHandler):

    def get(self, *args, **kwargs):
        # PXE: Get status snapshot for all installations
        self.logger.debug('%s.get() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        # Returns either installation.full_status(), or dict of uuid: installation.full_status()
        result = {}
        for uuid, inst in self.installations.iteritems():
            result[uuid] = inst.full_status()
        self.write(result)

    def post(self, *args, **kwargs):
        # PXE: Send response to installation status report
        self.logger.debug('%s.post() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        # Bundle request, there's a bunch of uuid=response fields in POST
        result = {'result': 'ok'}
        for uuid in self.request.body_arguments:
            if uuid not in self.installations:
                self.logger.error('StatusHandler().post(): Installation `%s` not found', uuid)
                result[uuid] = httplib.NOT_FOUND
                continue
            install = self.installations[uuid]
            if not install.waiting:
                self.logger.error(
                    'StatusHandler().post(): Installation `%s` not waiting for response', uuid)
                result[uuid] = httplib.REQUEST_TIMEOUT
                continue
            try:
                install.inbox.put_nowait(self.get_body_argument(uuid))
            except tornado.queues.QueueFull:
                self.logger.error(
                    'StatusHandler().post(): Installation `%s` already has response', uuid)
                result[uuid] = httplib.REQUEST_TIMEOUT
                continue
            result[uuid] = httplib.OK
        self.write(result)


class InstallationStatusHandler(SmartProxyHandler):

    def get(self, uuid, *args, **kwargs):
        # PXE: Get status snapshot for single installation
        self.logger.debug('%s.get(uuid=%r) args=%r, kwargs=%r',
                          type(self).__name__, uuid, args, kwargs)
        # Returns either installation.full_status(), or dict of uuid: installation.full_status()
        if uuid not in self.installations:
            return self.json_error(httplib.NOT_FOUND, 'Installation Not Found')
        self.write(self.installations[uuid].full_status())

    def post(self, uuid, *args, **kwargs):
        # PXE: Send response to installation status report
        self.logger.debug('%s.post(uuid=%r) args=%r, kwargs=%r',
                          type(self).__name__, uuid, args, kwargs)
        # May be a single or a bundle request.
        if uuid not in self.installations:
            return self.json_error(httplib.NOT_FOUND, 'Installation Not Found')
        if uuid in self.request.body_arguments:
            response = self.get_body_argument(uuid)
        elif 'response' in self.request.body_arguments:
            response = self.get_body_argument('response')
        else:
            return self.json_error(httplib.BAD_REQUEST, 'Invalid Request')
        install = self.installations[uuid]
        if not install.waiting:
            return self.json_error(httplib.REQUEST_TIMEOUT, 'Installation Not Waiting')
        try:
            self.installations[uuid].inbox.put_nowait(response)
        except tornado.queues.QueueFull:
            return self.json_error(httplib.REQUEST_TIMEOUT, 'Response Already Sent')
        self.write({'result': 'ok'})


class StatusFeedHandler(SmartProxyHandler):

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        # PXE: Long-polling for status change events
        self.logger.debug('%s.get() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        feed = {}
        try:
            event = yield self.pxe_feed.get(timeout=timedelta(seconds=10))
        except tornado.gen.TimeoutError:
            event = None
        while event is not None:
            self.pxe_feed.task_done()
            uuid = event.pop('uuid')
            if uuid not in feed:
                feed[uuid] = event
            else:
                inst = feed[uuid]
                if 'downloads' in event:
                    inst.setdefault('downloads', {}).update(event.pop('downloads'))
                inst.update(event)
            # Try to retrieve next event for next `while` iteration
            try:
                event = self.pxe_feed.get_nowait()
            except tornado.queues.QueueEmpty:
                event = None
        self.write(feed)


class StatusLogsHandler(SmartProxyHandler):

    def get(self, uuid, *args, **kwargs):
        # PXE: Download most recent installation logs from SmartProxy
        self.logger.debug('%s.get() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        if uuid not in self.installations:
            return self.json_error(httplib.BAD_REQUEST, 'Installation Not Found')
        inst = self.installations[uuid]
        if not inst.last_log:
            return self.json_error(httplib.NOT_FOUND, 'Installation Logs Not Found')
        path = os.path.join(settings.STORAGE_UUID, uuid, inst.last_log)
        with open(path, 'rb') as f:
            contents = f.read()
        timestamp = time.strftime('%Y-%m-%dT%H:%M:%S.00Z', time.gmtime(inst.logs[inst.last_log]))
        self.set_header('Last-Modified', timestamp)
        self.set_header('Content-Type', 'application/octet-stream')
        self.set_header('Content-Disposition', 'attachment; filename={}'.format(inst.last_log))
        self.write(contents)


class InstallationHandler(SmartProxyHandler):

    @tornado.gen.coroutine
    def post(self, uuid, *args, **kwargs):
        # Host action: send status update to SmartProxy and get response
        self.logger.debug('%s.post() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        if uuid not in self.installations:
            self.logger.error('Invalid installation requested: %s', uuid)
            self.set_status(httplib.BAD_REQUEST, 'Invalid Request')
            self.write('error')
            return
        status = self.get_body_argument('status')
        if status not in LEGAL_STATUS:
            self.logger.error('Illegal status reported by installation `%s`: %s', uuid, status)
            self.set_status(httplib.BAD_REQUEST, 'Invalid Request')
            self.write('error')
            return
        install = self.installations[uuid]
        if install.waiting:
            # Duplicate request from installation while there's another request waiting
            # Solution: fail request, attempt to fail other request
            self.logger.error('Status update by installation `%s` while in waiting mode', uuid)
            # Send 'error' response to existing request
            try:
                install.inbox.put_nowait('error')
            except tornado.queues.QueueFull:
                pass
            # Return 'error' to new request
            self.set_status(httplib.BAD_REQUEST, 'Invalid Request')
            self.write('error')
            return
        install.waiting = True
        event = install.update_status(status, self.request.remote_ip)
        self.pxe_feed.put_nowait(event)
        self.storage.save_installation(self.installations, install)
        try:
            # Wait for PXE response
            response = yield install.inbox.get(timeout=timedelta(seconds=10))
        except tornado.gen.TimeoutError:
            # If no response from PXE within timeout
            self.set_status(httplib.REQUEST_TIMEOUT, 'No Response From PXE')
            response = 'error'
        else:
            # If there's PXE response
            install.inbox.task_done()
            if response == 'error':
                self.set_status(httplib.BAD_REQUEST, 'Rejected By PXE')
        finally:
            # Reset waiting flag in either situation
            install.waiting = False
        self.write(response)

    def delete(self, uuid, *args, **kwargs):
        # Delete installation and all it's files
        self.logger.debug('%s.delete() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        # Ensure installation exists
        if uuid not in self.installations:
            return self.json_error(httplib.NOT_FOUND, 'Installation Not Found')
        # Remove installation folder and all it's files
        path = os.path.join(settings.STORAGE_UUID, uuid)
        try:
            shutil.rmtree(path)
        except OSError:
            self.logger.exception('Failed to remove installation folder: %s', uuid)
        # Create a failure event in case installation is waiting for one
        install = self.installations[uuid]
        try:
            install.inbox.put_nowait('error')
        except tornado.queues.QueueFull:
            pass
        # Forget about installation
        del self.installations[uuid]
        self.write({'result': 'ok'})


def if_not_responded(func):
    def decorated(self, *args, **kwargs):
        if self.responded:
            return
        func(self, *args, **kwargs)


@tornado.web.stream_request_body
class FileHandler(SmartProxyHandler):

    uuid = None
    filepath = None
    abspath = None
    responded = False

    def prepare(self):
        '''
        Request initialization method. Called before request body is available.

        Used to validate the information available in the URI and headers and possibly return an
        error if there are any issues.

        :return: None, may set self.responded = True if an error response was generated.
        '''
        self.logger.debug('### prepare(method = %r, uri = %r, path = %r)',
                          self.request.method, self.request.uri, self.request.path)
        try:
            prefix, self.uuid, self.filepath = self.request.uri.strip('/').split('/', 2)
        except ValueError:
            self.logger.error('Illegal path specified for FileHandler: %r', self.request.path)
            self.responded = True
            return self.json_error(httplib.NOT_FOUND, 'Illegal Path Specified')
        if prefix != 'files':
            self.responded = True
            return self.json_error(httplib.BAD_REQUEST, 'Illegal URL prefix.')
        self.filepath = self.filepath.rstrip('/')
        if self.uuid == 'shared':
            data_root = settings.STORAGE_SHARED
        else:
            if self.uuid not in self.installations:
                self.responded = True
                self.logger.error('Requested installation %r not exists.', self.uuid)
                return self.json_error(httplib.NOT_FOUND, 'Installation Not Found')
            data_root = os.path.join(settings.STORAGE_UUID, self.uuid)
        if not path_validator(self.filepath):
            self.responded = True
            self.logger.error('Path %r is not legal file path.', self.filepath)
            return self.json_error(httplib.BAD_REQUEST, 'Illegal Path Specified')
        self.abspath = os.path.abspath(os.path.join(data_root, self.filepath))
        if os.path.exists(self.abspath):
            if self.request.method == 'POST':
                self.responded = True
                self.logger.warning('Cannot upload file if already exists: %s/%s',
                                    data_root, self.filepath)
                return self.json_error(httplib.CONFLICT, 'File Already Exists')
        else:
            if self.request.method == 'POST':
                dirname = os.path.dirname(self.abspath)
                if not os.path.exists(dirname):
                    try:
                        os.makedirs(dirname)
                    except OSError:
                        self.responded = True
                        self.logger.exception('Failed to create folder for uploaded file: %s/%s',
                                              data_root, self.filepath)
                        return self.json_error(httplib.INTERNAL_SERVER_ERROR,
                                               'Cannot Create Folder')
            else:
                self.responded = True
                return self.json_error(httplib.NOT_FOUND, 'File Not Found')

    def data_received(self, data):
        '''
        Stream processing method, Tornado will asynchronously feed chunks of data to this method
        as request body is arriving. Only required for POST request as all other requests do not
        contain a request body.

        :param data: yet another chunk of request's POST body.
        :return: None, may set self.responded = True if an error was generated.
        '''
        if self.responded:
            return
        self.logger.debug('### data_received(%d bytes)', len(data))
        try:
            with open(self.abspath, 'a+b') as f:
                f.write(data)
        except IOError:
            self.responded = True
            self.logger.exception('Error when writing to file %s', self.abspath)
            return self.json_error(httplib.INTERNAL_SERVER_ERROR, 'Cannot Create File')

    def head(self, *args, **kwargs):
        # PXE action: Check if file exists
        if self.responded:
            return
        self.logger.debug('%s.head() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        self.write({'result': 'ok'})

    def get(self, *args, **kwargs):
        # PXE action: Download file
        if self.responded:
            return
        self.logger.debug('%s.get() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        if os.path.isfile(self.abspath):
            if self.uuid == 'shared':
                data_root = settings.REDIRECT_SHARED
            else:
                data_root = settings.REDIRECT_UUID + format(self.uuid) + '/'
            self.set_header(settings.REDIRECT_HEADER, '{}{}'.format(data_root, self.filepath))
            filename = os.path.basename(self.filepath)
            self.set_header('Content-Disposition', 'attachment; filename={}'.format(filename))
        elif settings.DEBUG:
            files = os.listdir(self.abspath)
            if settings.STORAGE_FILENAME in files:
                files.remove(settings.STORAGE_FILENAME)
            self.write({'result': 'ok', 'contents': files})
        else:
            self.json_error(httplib.FORBIDDEN, 'Debug Mode Only')

    def post(self, *args, **kwargs):
        # PXE action: Upload file
        if self.responded:
            return
        self.logger.debug('%s.post() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        self.write({'result': 'ok'})

    def delete(self, *args, **kwargs):
        # PXE action: Delete file or folder
        if self.responded:
            return
        self.logger.debug('%s.delete() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        install = None
        if self.uuid != 'shared':
            install = self.installations[self.uuid]
        if os.path.isfile(self.abspath):
            os.remove(self.abspath)
        else:
            os.rmdir(self.abspath)
        if install:
            if self.filepath in install.files:
                del install.files[self.filepath]
            if self.filepath in install.logs:
                del install.logs[self.filepath]
                if self.filepath == install.last_log:
                    install.last_log = None
            self.storage.save_installation(self.installations, install)
        self.write({'result': 'ok'})


class InstallationFileHandler(SmartProxyHandler):

    def get(self, uuid, filepath, *args, **kwargs):
        # Host action: download installation file from SmartProxy
        self.logger.debug('%s.get() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        if uuid not in self.installations:
            return self.json_error(httplib.NOT_FOUND, 'Installation Not Found')
        install = self.installations[uuid]
        filepath = filepath.rstrip('/')
        if not path_validator(filepath):
            return self.json_error(httplib.BAD_REQUEST, 'Illegal Path Specified')
        path = os.path.join(settings.STORAGE_UUID, uuid, filepath)
        if not(os.path.exists(path) and os.path.isfile(path)):
            return self.json_error(httplib.NOT_FOUND, 'File Not Found')
        self.set_header(settings.REDIRECT_HEADER,
                        '{}{}/{}'.format(settings.REDIRECT_UUID, uuid, filepath))
        self.set_header('Content-Disposition',
                        'attachment; filename={}'.format(os.path.basename(path)))
        self.pxe_feed.put_nowait(install.update_downloads(filepath, self.request.remote_ip))


class InstallationLogsHandler(SmartProxyHandler):

    def post(self, uuid, *args, **kwargs):
        # Host action: upload logs file to SmartProxy
        self.logger.debug('%s.post() args=%r, kwargs=%r', type(self).__name__, args, kwargs)
        if uuid not in self.installations:
            return self.json_error(httplib.NOT_FOUND, 'Installation Not Found')
        if 'logfile' not in self.request.files:
            return self.json_error(httplib.BAD_REQUEST, 'Logfile Not Provided')
        install = self.installations[uuid]
        fileobj = self.request.files['logfile'][0]
        with open(os.path.join(settings.STORAGE_UUID, uuid, fileobj.filename), 'w+b') as f:
            f.write(fileobj.body)
        self.pxe_feed.put_nowait(install.update_logs(fileobj.filename, self.request.remote_ip))
        self.storage.save_installation(self.installations, install)
        self.write({'result': 'ok'})


def parse_command_line_options():
    parser = optparse.OptionParser()

    parser.add_option(
        '-c', '--config',
        default=os.environ.get('SMARTPROXYCONFDIR', '.:/etc/smart-proxy/'),
        help='Specify location of config file (defaults to %SMARTPROXYCONFDIR%)',
    )

    parser.add_option(
        '-l', '--log', default=None,
        help='Specify location for script logs (stderr by default)'
    )

    parser.add_option(
        '-v', '--verbosity', default=0, action='count',
        help='Increase logging verbosity (stacks up to -vvvv)',
    )

    options, args = parser.parse_args()

    if not options.log or (options.log.lower() == 'none'):
        options.log = None

    return options


def main():
    options = parse_command_line_options()
    logging.basicConfig(
        filename=options.log,
        level=logging.CRITICAL - options.verbosity * 10,
        format='%(levelname)s @ %(asctime)s - %(message)s',
        datefmt='%Y/%m/%d %H:%M:%S',
    )
    logging.captureWarnings(True)
    logger = logging.getLogger(__name__)

    settings.update(load_yaml('settings.yaml', options.config.split(':')))
    storage = InstallationStorage(logger)
    params = {
        'logger': logger,
        'storage': storage,
        'installations': storage.load_installations(),
        'pxe_feed': tornado.queues.Queue()
    }

    url_map = [
        (r'/status/feed/?', StatusFeedHandler, params),
        (r'/status/(?P<uuid>[0-9a-f\-]{7,36})/logs/?', StatusLogsHandler, params),
        (r'/status/(?P<uuid>[0-9a-f\-]{7,36})/?', InstallationStatusHandler, params),
        (r'/status/?', StatusHandler, params),
        (r'/files/.*', FileHandler, params),
        (r'/(?P<uuid>[0-9a-f\-]{7,36})/logs/?', InstallationLogsHandler, params),
        (r'/(?P<uuid>[0-9a-f\-]{7,36})/(?P<filepath>.*)', InstallationFileHandler, params),
        (r'/(?P<uuid>[0-9a-f\-]{7,36})/?', InstallationHandler, params),
        (r'/', RootHandler, params),
        (r'/.*', ErrorHandler, params),
    ]
    app = tornado.web.Application(url_map, debug=settings.DEBUG)
    app.listen(settings.LISTEN_PORT)
    tornado.ioloop.IOLoop.current().start()

if __name__ == "__main__":
    main()
