#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Skeleton for test suit at the moment.

Runs SmartProxy with test settings in child thread, receives it's stderr output line by line.

run scenario:
    cd ..
    python smartproxy.py --config ./test -v
    stderr -> stream, check for ERROR: and CRITICAL: messages
    terminate on test suite conclusion
'''

import httplib
import json
import os
import requests
import shutil
import subprocess
import threading
import time
import unittest
import uuid


class TestingEnvironment(object):
    """
    Context manager for testing environment. Runs SmaryProxy in a child process using custom
    configuration file and a temporary folder for data storage. Terminates subprocess once the
    testing suite is completed and removes the data storage folder.
    """
    def __enter__(self):
        os.makedirs('./storage/shared')
        TEST_DIR = os.path.dirname(os.path.abspath(__file__))
        BASE_DIR = os.path.dirname(TEST_DIR)
        self.proxy = subprocess.Popen(
            args=['python', 'smartproxy.py', '-vv', '--config', TEST_DIR],
            close_fds=True,
            cwd=BASE_DIR,
        )
        self.http = requests.Session()
        for i in xrange(5):
            try:
                r = requests.get('http://localhost:8888/status/')
                if r.status_code == 200:
                    break
            except Exception:
                pass
            time.sleep(0.5)
        else:
            raise Exception('Failed to start SmartProxy subprocess.')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.proxy.terminate()
        shutil.rmtree('./storage')


class AsyncPost(threading.Thread):
    """
    Asynchronous POST request. Used to test Host's status update handler which requires waiting
    for PXE's reaction.
    """

    exception = None
    text = None
    json = None
    status_code = 0

    url = None
    post_data = None

    def __init__(self, url, post_data=None):
        super(AsyncPost, self).__init__()
        self.url = url
        self.post_data = post_data

    def run(self):
        try:
            r = requests.post(self.url, self.post_data)
        except Exception as e:
            self.exception = e
        else:
            self.status_code = r.status_code
            self.text = r.text
            try:
                self.json = json.loads(self.text)
            except Exception as e:
                pass


class TestStandardProcess(unittest.TestCase):

    http = None

    def assertRequest(self, url, method='GET', post_data=None, status_code=httplib.OK):
        try:
            request_method = getattr(self.http, method.lower())
        except AttributeError:
            self.fail('Unknown request method: `%s`' % method)
        if post_data:
            response = request_method(url, post_data)
        else:
            response = request_method(url)
        if status_code is not None:
            self.assertEquals(response.status_code, status_code)
        return response

    def assertJSONRequestEquals(self, url, method='GET', post_data=None,
                                status_code=httplib.OK, etalone=None):
        response = self.assertRequest(url, method, post_data, status_code)
        if etalone is not None:
            try:
                json_response = json.loads(response.text)
            except Exception as e:
                self.fail('Cannot parse response as JSON: %r\n%s' % (response.text, e))
            self.assertEquals(json_response, etalone)

    def assertJSONResponse(self, r, status_code, etalone):
        if status_code is not None:
            self.assertEquals(r.status_code, status_code)
        if etalone is not None:
            response = json.loads(r.text)
            self.assertEquals(response, etalone)
        return r

    def test_standard_process(self):
        with TestingEnvironment() as env:
            self.http = env.http

            # Ensure status is empty at the start.
            self.assertJSONRequestEquals('http://localhost:8888/status/', etalone={})

            # Create a new installation.
            uuid1 = str(uuid.uuid4())
            self.assertJSONRequestEquals('http://localhost:8888/', method='POST',
                                         post_data={'uuid': uuid1}, etalone={'result': 'ok'})

            # Check new installation is present in full status data.
            r = self.assertRequest('http://localhost:8888/status/')
            data = json.loads(r.text)
            self.assertIn(uuid1, data)

            # Upload shared file and check it's download headers are set correctly.
            test_data = 'test data'
            self.assertJSONRequestEquals('http://localhost:8888/files/shared/test1.txt',
                                         method='POST',
                                         post_data=test_data,
                                         etalone={'result': 'ok'})
            r = self.assertRequest('http://localhost:8888/files/shared/test1.txt')
            self.assertIn('X-Accel-Redirect', r.headers)
            self.assertEquals(r.headers['X-Accel-Redirect'], '/proxy_storage_shared/test1.txt')

            # Upload installation file and check it's download headers are set correctly.
            self.assertJSONRequestEquals('http://localhost:8888/files/%s/test2.txt' % uuid1,
                                         method='POST',
                                         post_data=test_data,
                                         etalone={'result': 'ok'})
            r = self.assertRequest('http://localhost:8888/%s/test2.txt' % uuid1)
            self.assertIn('X-Accel-Redirect', r.headers)
            self.assertEquals(r.headers['X-Accel-Redirect'],
                              '/proxy_storage_uuid/%s/test2.txt' % uuid1)

            # Ensure downloaded file is properly registered in the PXE events feed.
            r = self.assertRequest('http://localhost:8888/status/feed')
            data = json.loads(r.text)
            self.assertIn(uuid1, data)
            self.assertIn('downloads', data[uuid1])
            self.assertIn('test2.txt', data[uuid1]['downloads'])

            # Check status handling for host and PXE.

            # Ensure status update query does not return a result immediately.
            status_request = AsyncPost('http://localhost:8888/%s' % uuid1,
                                       {'status': 'sign_request'})
            status_request.start()
            time.sleep(0.1)
            self.assertTrue(status_request.is_alive())

            # Retrieve result via PXE feed.
            r = self.assertRequest('http://localhost:8888/status/feed')
            data = json.loads(r.text)
            self.assertIn(uuid1, data)
            self.assertIn('status', data[uuid1])
            self.assertIn('sign_request', data[uuid1]['status'])

            # Post PXE reaction to host's status update
            self.assertJSONRequestEquals('http://localhost:8888/status/%s/' % uuid1, method='POST',
                                         post_data={uuid1: 'ok'}, etalone={'result': 'ok'})

            # Receive and process PXE reaction on host's side
            status_request.join(timeout=0.5)
            self.assertFalse(status_request.is_alive())
            self.assertEquals(status_request.status_code, httplib.OK)


if __name__ == '__main__':
    unittest.main()
